require(grDevices)

downsample <- function(x, rres,cres,
  ridx = 1:nrow(x), cidx = 1:ncol(x), rcenter=TRUE, ccenter=TRUE)
{
  m <- length(ridx); n <- length(cidx);
  my <- (m+rres-1) %/% rres; ny <- (n+cres-1) %/% cres; 
  r <- .C("R_downsample",
    as.integer(nrow(x)),as.integer(ncol(x)),as.double(x),
    as.integer(rres),as.integer(cres),
    as.integer(m), as.integer(n),
    as.integer(ridx-1), as.integer(cidx-1),
    as.integer( rcenter + 2*ccenter ),
    y = double(my*ny), meany = double(1), sdy = double(1),
    NAOK=TRUE,DUP=FALSE);
  dim(r$y)=c(my,ny);
  attr(r$y,"mean") <- r$meany
  attr(r$y,"sd") <- r$sdy
  r$y
}

labelpack <- function(x,space = NULL, spacefrac = 0.05,
  minx = min(x), maxx = max(x), maxiter = 50){
  n <- length(x)
  o <- order(x)

  if(missing(space))
    space <- spacefrac*(maxx-minx)
    
  r <- .C("R_labelpack",as.integer(n),as.double(x[o]),
    as.double(space),as.double(minx),as.double(maxx),
    as.integer(maxiter),y = double(n))
  r$y[o] <- r$y
  r$y }

# red-black-green gradient
rg.colors <- c(rgb(0,seq(1,.1,-.1),0), rgb(0,0,0), rgb(seq(.1,1,.1),0,0) ) 

# red-white-blue gradient
rb.colors <- c(hsv(0.66,seq(1,.1,-.1),1), rgb(1,1,1), hsv(0,seq(.1,1,.1),1) ) 

# black-and-white
bw.colors <- gray(seq(21,0,-1)/21)

coldmap <- function(..., clust=NULL,
  dlab=NULL,
  clab = NULL,
  col.clab=NULL,
  cex.clab=.8,
  rannot = NULL,
  col=rb.colors,
  saturation=.666,
  downsampling = NULL,
  wa=NULL, wb=NULL,
  chang = 0.0,
  rhang = 0.05,
  gutter=0.5,
  rdw = 3, cdw =3
  )
{
  d <- list(...)
  K <- length(d)
  if(is.null(clust))
    clust <- nclust2(...)


  par0 <- par(no.readonly=TRUE)
  on.exit(par(par0))

  L <- array(c(0,seq(1,2*K,2),0,2*K+2,seq(2,2*K,2),2*K+1,rep(0,K+2)), dim=c(K+2,3))
  for(k in 1:K)
    L[1+k,3] = 2*K+2+k

  layout(
    L,
    widths=c(lcm(rdw),10,lcm(1.5)), heights=c(lcm(2),rep(1,K),lcm(cdw))  )

  # make per dataset ordering
  nrow.d <- sapply(d,function(u)nrow(u))
  b.row <- cumsum(nrow.d)
  a.row <- c(0,b.row[-K])+1
  N <- sum(nrow.d)

  # draw individual heatmap
  cc <- clust$cclust
  nu <- nrow(cc$unicol)
  if( is.null(wa) ) wa <- 1
  if( is.null(wb) ) wb <- nu

  if( is.null(downsampling[2]) )
    cres <- (wb-wa+1) %/% 400 + 1
  else
    cres <- downsampling[2]

  # get the total range of row dendrogram by constructing the total
  rc <- clust$rclust 
  rdend <- dendrolines(rc,hang=rhang,height=-rc$branch.level)
  rdend.range <- range(-rdend$y)
  
  for(k in 1:K) 
    {
    # row dendrogram
    subclust <- rc$subclust[[k]]
    u <- dendrolines(subclust,hang=rhang,
          height=-rc$branch.level[subclust$sup.branch])
    par(mar=c(gutter,2,gutter,1))
    plot(-u$y,u$x,type="l",axes=FALSE,ann=FALSE,
      ylim=c(.5,nrow.d[k]+.5), xlim=rdend.range, yaxs="i" )

    # dataset label
    mtext(dlab[k],side=2,line=1,at=nrow.d[k]/2,cex=.8)
    
    # heatmap
    par(mar=c(gutter,1,gutter,1))
    ik <- cc$unicol[,k]
    ik <- ifelse(ik == 0, NA, ik)

    if( is.null( downsampling[1] ) )
      rres <- nrow(d[[k]]) %/% 200 + 1
    else
      rres <- downsampling[1]

    dx <- t(downsample(
      d[[k]][ subclust$order, ik[cc$order[wa:wb]] ],
      rres, cres, 
      rcenter=FALSE,ccenter=TRUE ))
    
    image( tanh( saturation*(dx)/attr(dx,"sd")),
      zlim=c(-1,1),col=col,
      y=seq(1,nrow(d[[k]]),rres), ylab="",
      x=seq(wa,wb,cres), xlab="",
      axes=FALSE)
    axis(2,cex.axis=.8)
    if(k == K) axis(1,cex.axis=.8)
    
    }

  # column dendrogram
  h <- cc$branch.corr
  hcut <- cres
  for(i in length(h):1)
    {
    if(cc$merge[i,1] > 0 && cc$n.leaf[cc$merge[i,1]] <= hcut)
      h[cc$merge[i,1]] <- h[i]
    if(cc$merge[i,2] > 0 && cc$n.leaf[cc$merge[i,2]] <= hcut )
      h[cc$merge[i,2]] <- h[i]
    }
  u <- dendrolines(cc, hang=chang, height=-h)
  par(mar=c(1,1,2,1))
  plot(u$x,-u$y,type="l",axes=FALSE,ann=FALSE,xlim=c(wa-.5,wb+.5),
    xaxs = "i" )
  axis(2,at=c(0,.2,.4,.6,.8,1),labels=c("0",".2",".4",".6",".8","1"),cex.axis=.8)

  # gene labels
  par(mar=c(0,1,0,1))
  plot.new()
  plot.window(c(wa-.5,wb+.5),c(0,1),xaxs="i")
  if( !missing(clab) )
    {
    ordlab <- cc$labels[cc$order]
    if(is.character(clab))
      i <- unique(as.vector(na.omit(match( clab, ordlab ))))
    else
      i <- as.integer(clab)

    i <- i[ i >= wa & i <= wb ]

    if(length(i) > 0)
      {
      recordGraphics(
        {
        s <- 0.6*par("cin")[2] / par("pin")[1] * (wb-wa+1) * cex.clab
        j <- labelpack(i, space = s,
          minx = wa - 0.5 + .5*s , maxx = wb + .5 -.5*s  )
        text( x=j, y=rep(0.17, length(i)), cex = cex.clab,
          labels=lab, col=col.lab, adj=c(0,.5), srt=90,xpd=NA )

        # bent ticks
        apply(rbind(i,j),2,
          function(u)lines( c(u[1],u[1],u[2],u[2]), c(0,.05,.1,.15) ))
        },
        list(i=i,nu=nu,cex.clab=cex.clab,lab=ordlab[i],
        col.lab=col.clab,wa=wa,wb=wb),
        getNamespace("graphics")
        )
      }
    }

  # additional matrix
  if( !is.null(rannot) )
    for(k in 1:K)
      {
      if(!is.null(rannot) && !is.null(rannot[[k]]))
        {
        cpos <- unique(sort(c(rannot[[k]])))
        cid <- 1:length(cpos)
        names(cid) <- cpos
        z <- apply( t(rannot[[k]][clust$rclust$subclust[[k]]$order,]), c(1,2), function(u)cid[u] )
        par(mar=c(gutter,1,gutter,1))
        image(z=z, col=cpos,xaxs="i",yaxs="i",axes=F)
        }
      }
  # else, we should draw nothing

  invisible(clust)
}

coldmap.old <- function( x,
  clust = NULL,
  cclust = NULL,
  rclust = NULL,
  ccenter = TRUE,
  rcenter = FALSE,
  downsampling = NULL,
  col = rg.colors,
  saturation = .666, 
  clab = NULL, 
  labspace=0.01,
  labcol = NULL,
  zc = NULL,
  zw = NULL
  ){
  m <- dim(x)[1]
  n <- dim(x)[2]

  # use supplied clustering
  if( class(clust) == "list"
    && class(clust$cclust) == "hclust"
    && class(clust$rclust) == "hclust" )
    { cclust <- clust$cclust; rclust <- clust$rclust; }

  # or compute when not available
  if( class(cclust) != "hclust")
    cclust <- nclust(x,impute=TRUE)
  if( class(rclust) != "hclust")
    rclust <- nclust(t(mostvar(x,n %/% 2)),center="both")

  if( length(cclust$order) != n || length(rclust$order) != m )
    stop("clust objects size doesn't match that of the matrix")

  # ordered labels
  ordlab <- cclust$labels[cclust$order]

  # determine the window
  if(is.null(zc))
    { wa <- 1; wb <- n; zw <- n }
  else
    {
    if(is.character(zc))
      zc <- match(zc, ordlab)
    if(is.na(zc) || is.null(zw) )
      { wa <- 1; wb <- n; zw <- n }
    else
      {
      wa <- max( 1, zc - zw %/% 2 )
      wb <- min( n, zc + zw %/% 2 )
      }
    }

  ## start drawing
  default.par <- par(no.readonly=TRUE)

  layout( array(c(0,2,0,1,3,0,0,4,0),dim=c(3,3)),
    widths=c(lcm(4),1,lcm(2.5)),
    heights=c(lcm(3),1,lcm(1))) 

  # top dendrogram
  v <- dendrolines(rclust)
  par(mar=c(2,0,0,0))
  plot(v$x,v$y,type="l",axes=FALSE,ann=FALSE, xlim=c(.5,m+.5), xaxs = "i")

  # left dendrogram
  u <- dendrolines(cclust,hang=0.025)
  par(mar=c(0,0,0,2))
  plot(-u$y,u$x,type="l",axes=FALSE,ann=FALSE, yaxs="i", ylim=c(wa,wb))

  # heatmap
  par(mar=c(0,0,0,0))

  if( is.null(downsampling) )
    { rds <- m %/% 500 + 1; cds <- zw %/% 500 + 1; }
  else
    { rds <- max(1,downsampling[1]); cds <- max(1,downsampling[2]) }
  
  dx <- downsample( x,
    rds, cds, ridx = rclust$order, cidx = cclust$order[wa:wb],
    rcenter = rcenter, ccenter = ccenter )

  image( tanh(saturation*(dx-attr(dx,"mean"))/attr(dx,"sd")),
    col=col,zlim=c(-1,1),
    y=seq(wa,wb,cds), ylab="",
    x=seq(1,m,rds), xlab="",axes=FALSE)
  axis(2)
  axis(3)

  # column labels
  if( !missing(clab) )
    {
    if(is.character(clab))
      i <- unique(as.vector(na.omit(match( clab, ordlab ))))
    else
      i <- as.integer(clab)

    i <- i[i >= wa & i <= wb]
    plot.new()
    plot.window(c(0,1),c(wa-.5,wb+.5),yaxs="i")
    space <- labspace*zw
    if(length(i) > 0)
      {
      j <- labelpack(i, space = space,
        minx = wa - 0.5 + .5*space, maxx = wb + .5 -.5*space )
      text( rep(0.205, length(i)), j, ordlab[i],
        pos = 4, col = labcol )

      # bent ticks
      apply(rbind(i,j),2,
        function(u)lines(c(0,.05,.1,.2),c(u[1],u[1],u[2],u[2])))
      }
    }
  invisible(list(rclust = rclust,cclust = cclust))
  }
